// Diafaan SMS Server Scripting Gateway script
// Clickatell HTTP Callback handling for received messages
// Last update 20-01-2020 14:15
// AddReference System.dll
// AddReference System.Web.dll
// AddReference HttpServer.dll
//
using System;
using System.Collections;
using System.Collections.Generic;
using System.Net;
using System.Web;
using System.Text;
using System.IO;
using HttpServer;

namespace DiafaanMessageServer
{
	public class GatewayScript : IScript
	{
		private int httpServerPort = 80;
		private int httpServerBacklog = 100;
		private IScriptHost host = null;
  
		public void OnLoad(IScriptHost host)
		{
			this.host = host;
			StartHttpServer();
		}
		public void OnUnload()
		{
			StopHttpServer();
		}
		#region Clickatell specific functions
		private string HandleClickatellQuery(Dictionary<string, string> queryList)
		{
			string response = "OK";
			string messageId = "";
			string toAddress = "";
			string fromAddress = "";
			string characterSet = "";
			string messageText = "";
			string pdu = "";
			string timeStamp;
			DateTime sendTime = DateTime.Now;
			DateTime receiveTime = DateTime.Now;
			string messageType = "sms.text";
			
			try {
				// Get the message parameters from the Clickatell query string
				messageId = HttpUtility.UrlDecode(GetQueryValue(queryList, "moMsgId"));			
				toAddress = HttpUtility.UrlDecode(GetQueryValue(queryList, "to"));	
				fromAddress = HttpUtility.UrlDecode(GetQueryValue(queryList, "from"));
				characterSet = HttpUtility.UrlDecode(GetQueryValue(queryList, "charset"));			
				messageText = HttpUtility.UrlDecode(GetQueryValue(queryList, "text"));
				if (characterSet.ToUpper() == "UTF-16BE") {
					messageType = "sms.unicode";
					try {
						messageText = HttpUtility.UrlDecode(GetQueryValue(queryList, "text"), Encoding.BigEndianUnicode);
					}
					catch (Exception e) {
						PostEventLog(e.Message, e.ToString(), EventLog.Error);					
					}
				}
				else if (characterSet.ToUpper() == "8-BIT") {
					messageType = "sms.binary";
					string udh = HttpUtility.UrlDecode(GetQueryValue(queryList, "udh"));
					if (udh.Length > 0) messageText = udh + "," + messageText;
				}
				pdu = HttpUtility.UrlDecode(GetQueryValue(queryList, "pdu"));
				timeStamp = HttpUtility.UrlDecode(GetQueryValue(queryList, "timestamp"));
				if (timeStamp.Length == 18) {
					try {
						sendTime = DateTime.ParseExact(timeStamp + " +2", "yyyy-MM-ddHH:mm:ss z", null);
					}
					catch (Exception e) {
						PostEventLog(e.Message, e.ToString(), EventLog.Error);					
					}
				}
			}
			catch (Exception e) {
				PostEventLog(e.Message, e.ToString(), EventLog.Error);					
			}
			// Add the received message to the message queue of Diafaan SMS Server
			if (pdu != "") {
				PostReceiveMessage(toAddress, fromAddress, messageText, messageType, messageId, sendTime, receiveTime, pdu);
			}
			return response;
		}
		public static byte[] EncodeHexStringToByteArray(string hexString)
		{
			byte[] returnValue = new byte[0];
			int packetLength;
			int i;

			packetLength = hexString.Length / 2;
			returnValue = new byte[packetLength];
			for (i = 0; i < packetLength; i++) {
				returnValue[i] = Convert.ToByte(Convert.ToByte(hexString.Substring(2 * i, 2), 16));
			}
			return returnValue;
		}
		#endregion
		
		#region HTTP server functions
		private HttpServer.HttpListener httpListener = null;
		private bool StartHttpServer()
		{
			bool returnValue = false;
			
			try {
				httpListener = HttpServer.HttpListener.Create(IPAddress.Any, httpServerPort);
				httpListener.RequestReceived += OnRequest;
				httpListener.Start(httpServerBacklog);
				PostEventLog(String.Format("Clickatell HTTP Callback Server started on host {0}, port {1}.", Dns.GetHostName(), httpServerPort), "", EventLog.Information);
				returnValue = true;
			}
			catch (Exception e) {
				PostEventLog(e.Message, e.ToString(), EventLog.Error);					
			}
			return returnValue;
		}
		private void StopHttpServer()
		{
			try {
				httpListener.Stop();
				PostEventLog(String.Format("HTTP Callback Server stopped."), "", EventLog.Information);
			}
			catch (Exception e) {
				PostEventLog(e.Message, e.ToString(), EventLog.Error);					
			}
		}
		private void OnRequest(object source, RequestEventArgs args)
		{
			IHttpClientContext context = (IHttpClientContext)source;
			IHttpRequest request = args.Request;
			Dictionary<string, string> queryList;
			string body = "ERROR";
			long length;

			try {
				if (request.BodyIsComplete && request.Method.ToUpper().Trim() == "POST") {
					queryList = ParseQueryString(GetHttpPostQueryString(request));
					queryList.Add("pdu", GetHttpPostQueryString(request));
				}
				else {
					queryList = ParseQueryString(request.Uri.Query.TrimStart('?'));
					queryList.Add("pdu", request.Uri.Query);
				}
				body = HandleClickatellQuery(queryList);
				length = Encoding.UTF8.GetBytes(body).Length;
				string response = string.IsNullOrEmpty(body)
								? "HTTP/1.1" + " " + (int) HttpStatusCode.OK + " " + "" + "\r\n\r\n"
								: string.Format("{0} {1} {2}\r\nContent-Type: {5}\r\nContent-Length: {3}\r\n\r\n{4}",
											"HTTP/1.1" , (int) HttpStatusCode.OK , "" ?? HttpStatusCode.OK.ToString(),
											length, body, "text/plain");
				byte[] buffer = Encoding.UTF8.GetBytes(response);
				context.Send(buffer);
			}
			catch (Exception e) {
				PostEventLog(e.Message, e.ToString(), EventLog.Error);					
			}
			finally {
				try {
					context.Close();
				}
				catch (Exception e) {
					PostEventLog(e.Message, e.ToString(), EventLog.Error);					
				}
			}
		}
		private Dictionary<string, string> ParseQueryString(string query)
		{
			Dictionary<string, string> returnValue = new Dictionary<string, string>();
		
			string[] parameterList = query.Split(new char[] {'&'}, StringSplitOptions.RemoveEmptyEntries);
			foreach (string parameter in parameterList) {
				string[] field = parameter.Split('=');
				if ((field.Length >= 2) && (!returnValue.ContainsKey(field[0]))) {
					returnValue.Add(field[0], field[1]);
				}
			}
			return returnValue;
		}
		private string GetHttpPostQueryString(IHttpRequest httpListenerRequest)
		{
			string returnValue = "";
			StreamReader reader;

			try {
				reader = new StreamReader(httpListenerRequest.Body, Encoding.UTF8);
				returnValue = reader.ReadToEnd();
			}
			catch (Exception e) {
				PostEventLog(e.Message, e.ToString(), EventLog.Error);					
			}
			return returnValue;
		}
		private string GetQueryValue(Dictionary<string, string> query, string name)
		{
			if (query.ContainsKey(name)) {
				return query[name];
			}
			else {
				return "";
			}
		}
		
		#endregion
		
		private void OnMessageToSend(string recordId, string toAddress, string fromAddress, 
		                             string message, string messageType)
		{
			PostSendResult(recordId, "", StatusCode.SendError, "Error: This Scripting Gateway does not have functionality to send messages.", "", "", false);
		}
		private void OnStatusRequest(string messageId)
		{
			// TODO: Send status request query
		}
		#region Gateway Script helper functions
		private enum StatusCode {Sent=200, Received=201, SendError=300}
		private enum EventLog {Information=0, Warning=1, Error=2}
		private enum TraceLog {Information=0, Send=1, Receive=2}

		public void OnMessagePacket(Hashtable messagePacket)
		{
			if (GetPacketString(messagePacket, "PacketName", "") == "MessageOut") {
				OnMessageToSend(GetPacketString(messagePacket, "RecordId", ""),
				                GetPacketString(messagePacket, "To", ""),
				                GetPacketString(messagePacket, "From", ""),
				                GetPacketString(messagePacket, "Message", ""),
				                GetPacketString(messagePacket, "MessageType", ""));
			}
			else if (GetPacketString(messagePacket, "PacketName", "") == "StatusRequest") {
				OnStatusRequest(GetPacketString(messagePacket, "MessageId", ""));
			}
		}
		private void PostReceiveMessage(string toAddress, string fromAddress, string messageText,
		                                string messageType, string messageId, DateTime sendTime,
		                                DateTime receiveTime, string Pdu)
		{
			Hashtable messagePacket = new Hashtable();
			
			messagePacket.Add("PacketName", "MessageIn");
			messagePacket.Add("To", toAddress);
			messagePacket.Add("From", fromAddress);
			messagePacket.Add("Message", messageText);
			messagePacket.Add("MessageType", messageType);
			messagePacket.Add("MessageId", messageId);
			messagePacket.Add("SendTime", sendTime.Ticks.ToString());
			messagePacket.Add("ReceiveTime", receiveTime.Ticks.ToString());
			messagePacket.Add("PDU", Pdu);
			host.PostMessagePacket(messagePacket);			
		}
		private void PostSendResult(string recordId, string messageId, StatusCode status,
		                            string statusText, string errorCode, string errorText, bool sendStatusRequest)
		{
			Hashtable messagePacket = new Hashtable();
			
			messagePacket.Add("PacketName", "MessageOutResult");
			messagePacket.Add("RecordId", recordId);
			messagePacket.Add("MessageId", messageId);
			messagePacket.Add("StatusCode", ((int) status).ToString());
			messagePacket.Add("StatusText", statusText);
			messagePacket.Add("ErrorCode", errorCode);
			messagePacket.Add("ErrorText", errorText);
			if (sendStatusRequest) {
				messagePacket.Add("SendStatusRequest", "1");
			}
			else {
				messagePacket.Add("SendStatusRequest", "0");
			}
			host.PostMessagePacket(messagePacket);			
		}
		private void PostEventLog(string eventMessage, string eventDescription, EventLog eventType)
		{
			Hashtable messagePacket = new Hashtable();
			
			messagePacket.Add("PacketName", "EventLog");
			messagePacket.Add("EventType", ((int) eventType).ToString());
			messagePacket.Add("EventMessage", eventMessage);
			messagePacket.Add("EventDescription", eventDescription);
			host.PostMessagePacket(messagePacket);			
		}
		private void PostTraceLog(string traceLogMessage, TraceLog logType)
		{
			Hashtable messagePacket = new Hashtable();
			
			messagePacket.Add("PacketName", "TraceLog");
			messagePacket.Add("TraceLogType", ((int) logType).ToString());
			messagePacket.Add("TraceLogMessage", traceLogMessage);
			host.PostMessagePacket(messagePacket);			
		}
		private string GetPacketString(Hashtable hashtable, string key, string defaultValue)
		{
			try {
				return (string) hashtable[key];
			}
			catch {}
			return defaultValue;
		}
		#endregion
	}
}
